/*       -*- C++ -*- */
/** \file
********************************************************************
* $URL: svn+ssh://nalw02/home/alg/schlemme/svn/cextools/trunk/include/cexfile/concept.h $
* $Id: concept.h 15 2010-06-25 19:53:27Z schlemme $
* $Date: 2010-06-25 21:53:27 +0200 (Fr, 25. Jun 2010) $
* $Revision: 15 $
* \author Tobias Schlemmer <keinstein@users.berlios.de>
* $Author: schlemme $ 
* \license LGPL
* 
* Cex file concept class. 
*
********************************************************************
*/	  
#ifndef CEXFILE_CONCEPT_H
#define CEXFILE_CONCEPT_H

#include "cexfile/concept_base.h"
#include "assert.h"
#include "cexfile/object.h"



namespace cexfile {
	/// This class describes a concept.
	class concept : public concept_base {
	protected:

		/// the objects, whose object concept we are
		object_rs objects;

		/// the extent of the concept
		object_rs extent;

		/// attributes, whose attibute conexpt we are
		intent attr;

	public:
		/// default construcotr
		concept():concept_base() {}
    
		/// standard consturctor
		/** \param Node root node of the concept */
		concept(const xmlpp::Node * Node):concept_base(Node) {
			attr=attributes;
		}

		/// creates a pseudo concept
		/** this constructor is mainly inteded to be used for comparisons
		    between concepts and intents.
		    \param i intent, which our concept will have;
		*/
		concept(const intent & i):concept_base() {
			attributes = i;
		}

		/// addes an objcet to the object (concept) list
		/** \param o object to be added 
		 */
		void add_object(const object & o) {
			objects += o;
		}

		/// sets the set of all objects (object concepts)
		/** \param s set of objects */
		void set_objects (const object_rs & s) { objects = s; }

		/// return the set of objects (object concepts)
		const object_rs * get_objects () const { return &objects; }

		/// return the number of objects (object concepts)
		int get_object_count() const { return objects.size(); }

		/// return the number of attributes (attribute concepts)
		int get_attribute_count() const { return attr.size(); }

		/// set the attribute (concept) set
		/** \param i set of attributes */
		void set_attributes(intent & i) { attr = i; } 

		/// Return the set of Attributes
		/** This function returns the set of attributes, whose attribute
		    concepts the current concept is.
		    \return attribute set
		*/
		const intent & get_attributes () const { return attr; }

		/// return the number of attributes in the intent
		intent::size_type intentsize() { return attributes.size(); } 

		/// compare linearly with another concept
		/** \param c concept to be compared with
		    \retval true we are below the concept
		    \retval false otherwise
		*/
		bool operator < (const concept & c) const{
			return (concept_base) *this < (concept_base) c;
		}

		/// compare linearly with an object (concpt)
		/** \param o object to be compared with 
		    \retval true we are below the corresonding object concept 
		    \retval false otherwise
		*/
		bool operator < (const object & o) const {
			return (object_base) *this < (object_base) o;
		}

		/// add an object to the list of objects (object concept)
		/** \param o object to be added
		    \return self reference 
		*/
		concept & operator += (const object & o) {
			objects += o;
		}

		/// subtracts an attribute set from the set of attributes
		/** \param i set of attributes
		    \return self reference 
		*/
		concept & operator -= (const intent & i) {
			attr -= i;
			return *this;
		}
	};

	/// compare linearly an object to a concept
	/** \relates object
	    \relates concept 
	    \param o object
	    \param c concept
	    \retval true the object concept of \p o is below the concept \p c
	*/
	inline bool operator< (const object & o, const concept & c) {
		return (object_base) o < (object_base) c;
	}
  
	/// set of concepts
	/* relates concept */
	typedef std::set<concept> concept_s;

	/// add a concept to a concept set
	/** \relates cexfile::concept_s
	    \relates cexfile::concept
	    \param s set to be appended to
	    \param c concept to be added
	    \return reference to the set \p s
	*/
	inline concept_s& operator += (concept_s& s, const concept &c) {
		s.insert(c);
		return s;
	}


	/// concept reference;
	typedef concept_s::iterator concept_r;

	/// concept reference vector.
	typedef std::vector<concept_r> concept_rs;

	/// add a concept to a concept set
	/** \relates cexfile::concept_rs
	    \relates cexfile::concept_r
	    \param s vector to be appended to
	    \param c concept reference to be added
	    \return reference to the set \p s
	*/
	inline concept_rs& operator += (concept_rs& s, const concept_r &c) {
		s.push_back(c);
		return s;
	}


	class concept_sibling;

	/// pointer to a concept sibling
	/** \relates concept_sibling */
	typedef concept_sibling* ref_concept_sibling;

	/// a set of pointers  to concept siblings
	/** \relates concept_sibling */
	typedef std::list<ref_concept_sibling> concept_sibling_rl;

	/// a node in the concept lattice order
	/** \todo Can this be merged with concept? */
	class concept_sibling:public concept_sibling_rl{
	private:
		/// the concept, we refer to
		const concept* ref;
	protected:
		/// lower neighbours
		concept_sibling_rl neighbours;
	public:
		friend class concept_order; 

		/// default constructor
		concept_sibling():ref(NULL) {}
    
		/// standard constructor
		/** \param r pointer to the concept we refer to 
		 */
		concept_sibling(const concept * r):ref(r) {
#ifdef DEBUG
			std::cerr << "Instantiating concept_sibling at position "
				  << ref->get_position() << std::endl;
#endif 
		}

		/// returns the corresponding concept
		/** \return concept, we refer to 
		 */
		concept_s::const_reference  get_ref() const {return *ref;}

		/// returns a pointer to our concept
		/** \return point to the concept, we refer to
		 */
		const concept *  get_refp() const {return ref;}

		/// returns the list of lower neighbours
		/** \return list of lower neighbours 
		 */
		const concept_sibling_rl &get_neighbours() const { return neighbours; }
  
		/// adds an lower concept_sibling
		/** this function adds a concept_sibling to the lists of 
		    lower concept_siblings and lower neighbours if appropiate.
		    \retval true we found the concept somewhere below us
		    \retval false the concept could not be added because its not <= us
		*/
		bool check_neighbours(ref_concept_sibling c) {
      
			if (back() == c) return true;
#ifdef DEBUG
			std::cerr << c->get_ref().get_intent() << " <= " 
				  << get_ref().get_intent() << "?" << std::flush;
#endif
			if (!(get_ref().get_intent() <= c->get_ref().get_intent())) {
#ifdef DEBUG
				std::cerr << " No." << std::endl;
#endif

				return false; 
			}
#ifdef DEBUG
			std::cerr << " Yes." << std::endl;
#endif
			bool found = false;
			for (concept_sibling_rl::iterator i=neighbours.begin();
			     i!=neighbours.end(); i++) {
				bool b = ((*i)->check_neighbours(c));
				found = b || found;
			}
			push_back(c);
			if (!found) {
#ifdef DEBUG
				std::cerr << "Added: " << c->get_ref().get_intent() 
					  << " < " << get_ref().get_intent()
					  << std::endl;
#endif
				neighbours.push_back(c);
			}
			return true;
		}
    
		/// collects the extent for our concept
		/** The object concepts must already be distributed.
		    \retval reference to the object list of our concept 
		*/
		const object_rs & check_objects() {
			object_rs & objects=*(const_cast<object_rs *>(ref->get_objects()));

			// already visited
			if (objects.size()) return objects;

			for (concept_sibling_rl::iterator i = neighbours.begin();
			     i != neighbours.end(); i++) {
				objects += (*i) -> check_objects();
			}
			return objects;
		}
    
		/// calculates the set of attributes, we belong to as a attribute concept
		/** The intents of the concepts must be set already. */
		void check_attributes () {
#ifdef DEBUG
			std::cerr << "Check attributes: " << ref->get_intent() << std::endl;
#endif
			for (concept_sibling_rl::iterator i = neighbours.begin();
			     i != neighbours.end(); i++) {
#ifdef DEBUG
				std::cerr << "removing from " 
					  << ((*i)-> ref -> get_attributes())
					  << "; " << ((*i) ->ref->get_intent())
					  << std::endl;
#endif
				*const_cast<concept *>((*i) ->ref) -= ref->get_intent();
#ifdef DEBUG
				std::cerr << "got " << (*i)->ref->get_attributes() << std::endl;
#endif
			}
		}

	};

	/// list of concept_sibling
	/** \relates concept_sibling
	 */
	typedef std::list<concept_sibling> concept_sibling_l;

	/// outer set of the concept lattice order
	class concept_order:public concept_sibling_l {
	public:
		/// default constructor
		concept_order():concept_sibling_l() {}

		/// standard constructor
		/** This constructor initializes the order according to a
		    concept set.
		    \param s set of concepts
		*/	
		concept_order(concept_s & s):concept_sibling_l() {
			set(s);
		}

		/// initializes the concept order from a given set  of concepts
		/** \param s set of concepts
		    \return self reference
		*/
		concept_order & set(concept_s & s) {
			const concept* old = NULL;
			for (concept_s::const_iterator i = s.begin(); i!=s.end(); i++) {
#ifdef DEBUG
				std::cerr << i->get_intent() << std::endl;
				if (old)
					if (!(*old < *i)) abort();
				old = & (*i);
#endif
				push_front(concept_sibling(&(*i)));
			}

			iterator i = begin();
			for (i++; i!=end(); i++) {
				front().check_neighbours(&(*i));
			}
			return *this;
		}
    
		/// initializes the extents of the concepts
		/** the object concepts must already be distributed
		 */
		void check_objects() {
			front().check_objects();
		}

		/// initializes the attribute sets of the concepts
		/** the intents of the concepts must already be set
		 */
		void check_attributes() {
			for (iterator i = begin(); i!= end(); i++)
				i->check_attributes();
		}
	};
}


#endif /* CEXFILE_CONCEPT_H */
