/*       -*- C++ -*- */
/** \file
********************************************************************
* $URL: svn+ssh://nalw02/home/alg/schlemme/svn/cextools/trunk/include/cexfile/diagram.h $
* $Id: diagram.h 15 2010-06-25 19:53:27Z schlemme $
* $Date: 2010-06-25 21:53:27 +0200 (Fr, 25. Jun 2010) $
* $Revision: 15 $
* \author Tobias Schlemmer <keinstein@users.berlios.de>
* $Author: schlemme $ 
* \license LGPL
* 
* Diagram class.
********************************************************************
*/	  
#ifndef CEXFILE_DIAGRAM_H
#define CEXFILE_DIAGRAM_H

#include <iostream>
#include <cstring>
#include <vector>
#include <cmath>
#include "cexfile/constants.h"
#include "cexfile/attribute_label.h"
#include "cexfile/object_label.h"
#include "cexfile/edge.h"

namespace cexfile {

	/// diagram class
	/** This class contains the translated structure of one diagram node.
	    It calculates all necessary data and links to draw the lattice.
	*/
	class diagram {
		/// XML node of the diagram
		const xmlpp::Node * node;
		/// List of contexts.
		/** \todo Check, how many contexts a diagram can have. Currently
		    Context explorer seems to support only one
		    context. Nevertheless the real context of a lattice can be a
		    subcontext, as some objects and attributes can be left out. */
		context_s contexts;
    
		/** \name named XML entities
		    These entities contain the unchanged XML attribute content. 
		    They are not decoded to any binary data.
		    \{ */
		/// describes, how attribute labels shall be constructed
		/** known values are:
		    \li NoAttribsLabelsStrategy
		    \li AllAttribsLabelsStrategy
		    \li AllAttribsMultiLabelsStrategy
		*/
		Glib::ustring attribute_label_display_mode;
    
		/// describes, how object labels shall be constructed
		/** known values are:
		    - NoObjectsLabelsStrategy
		    - AllObjectsLabelsStrategy
		    - AllObjectsMultiLabelsStrategy
		    - OwnObjectsCountLabelsStrategy
		    - ObjectsCountLabelsStrategy
		    - StabilityLabelsStrategy
		*/
		Glib::ustring object_label_display_mode;

		/// describes, how the node radius shall be calculated
		/** known values are:
		    - OwnObjectsVolumeNodeRadiusCalcStrategy
		    - MaxNodeRadiusCalcStrategy
		    - ObjectVolumeNodeRadiusCalcStrategy
		    - ConceptStabilityToDesctructionNodeRadiusCalcStrategy
		*/
		Glib::ustring node_radius_mode;

		/// describes, how edges shall be calculated
		/** known values are:
		    - FixedEdgeSizeCalcStrategy
		    - InvisibleEdgeStrategyModel
		    - ObjectFlowEdgeSizeCalcStrategy
		    - ConceptConnectionEdgeSizeCalcStrategy
		*/
		Glib::ustring edge_display_mode;

		/// describes what shall be highlighted if a node is selected
		/** known values are:
		    - FilterIdealHighlightStrategy
		    - OneNodeHighlightStrategy
		    - NeigboursHighlightStrategy
		    - IdealHighlightStrategy
		    - FilterHighlightStrategy
		    - NoHighlightStrategy
		*/
		Glib::ustring highlight_mode;

		/// describes which heuristic to use for calculation of concept positions
		/** known values are:
		    - MinIntersectionLayout
		    - ChainDecompositionLayout
		    - FreeseLayout
		    - ForceLayout
		*/
		Glib::ustring lattice_layout;
		/*\} */

		/// font size in cex units
		double label_font_size;

		/// (maximal) node radius
		double node_radius;

		/// minimal node radius
		double min_node_radius;

		/// grid size X coordinate
		int grid_size_x;

		/// grid size Y coordinate
		int grid_size_y;

		/// show collisions between edges and nodes
		bool show_collisions;

		/// maximal value to be assumed, when node radius depends on other
		/// quantities
		int node_radius_max_value;

		/// list of concepts
		concept_s concepts;

		/// list of (Multi-)labels below nodes (e.g. objects)
		label_s downlabels;

		/// list of (Multi-)labels above nodes (e.g. attributes)
		label_s uplabels;

		/// list of single attribute labels
		attribute_label_s attribute_labels;

		/// list of single object labels
		object_label_s object_labels;

		/// hasse relation
		concept_order order;
    
		/// edges in the diagram
		edge_s edges;
    
	public:
		/// default constructor
		diagram():node(NULL),min_node_radius(2) {}

		/// standard constructor
		/** This constructor reads a diagram from a XML structure.
		    \param Node root node of the diagram
		    \param c context on which the diagram is based
		*/
		diagram(const xmlpp::Node * Node, context_s& c);

		/// return font size
		/** This function returns the font size read from the cex file.
		    \return Font size in cex units
		*/
		double  get_font_size() const { return label_font_size; }

		/// return the list of edges
		/** \return list of edges */
		const edge_s & get_edges() const { return edges; }

		/// return the list of concepts 
		/** \return list of concepts */
		const concept_s & get_concepts() const { return concepts; }

		/// return the list of irreducebles
		/** \return list of references (iterators) to irreducebles */
		concept_rs get_irreducibles() const;

		/// return the context
		/** \return list of contexts */
		const context_s & get_contexts() const { return contexts; }

		/// return the context
		/** \return single context */
		const context & get_actual_context() const { 
			return (*(contexts.begin())).second; 
		}

		/// return the upper multilabels
		/** \return list of upper multilabels */
		const label_s & get_uplabels() const { return uplabels; }

		/// return the list of lower multilabels
		/** \return list of multilabels below nodes */
		const label_s & get_downlabels() const { return downlabels; }

		/// return the node radius calculation strategy
		/** \return name of the current node radius calculation strategy */
		const Glib::ustring & get_node_radius_mode() const 
			{ return node_radius_mode;}

		/// return the total number of objects
		/** \return total of objects of the context(s) */
		int get_object_count() const {
			int ret = 0;
			for (context_s::const_iterator i=contexts.begin();
			     i!= contexts.end(); i++) 
				ret += i->second.get_object_count();
			return ret;
		}

		/// calculates the node radius
		/** \param c concept for which the radius shall be calculated 
		    \return radius in cex units */
		double get_node_radius (const concept & c) const;

		/// calculates the node radius
		/** \param ratio ratio between actual/maximal value according to 
		    node calculation strategy
		    \return radius in cex units */
		double get_node_radius (double ratio) const {
#ifdef DEBUG
			std::cerr << "Node radius: (" 
				  << node_radius << " - " << min_node_radius
				  << ") * sqrt(" << ratio << ") + " 
				  << min_node_radius << std::endl;
#endif
			return (node_radius - min_node_radius) * sqrt(ratio) + min_node_radius;
		}

		/// load diagram settings form XML
		/** \param Node root node of the diagram */
		void load_settings(const xmlpp::Node * Node);
		/// finds the edges for the hasse diagram
		void find_edges();

		/// calculates and distributes the object concepts
		void set_concept_objects();

		/// calculates the extents of the concepts
		void set_concept_extents();

		/// calculates and distributes the attribute concepts
		void set_concept_attr();

		/// distributes the lower multilabels to the concepts
		void update_downlabels();
    
		/// distributes the upper multilabels to the concepts
		void update_uplabels();

		/*
		  bool has_attribute(const intent & i) const {
		  for (context_s::const_iterator it = contexts.begin();
		  it != contexts.end();
		  it++) 
		  if (it->second.has_attribute(i)) return true;
		  return false;
		  }
		*/

		/// some statistical output
		/** \param o output stream, where the statistics shall occure 
		    \return reference to the output stream */
		std::ostream & statistics(std::ostream & o) const {
			o << " Concepts: " << concepts.size() << std::endl;
			o << " down labels: " << downlabels.size() << std::endl;
			o << " up labels: " << uplabels.size() << std::endl;
			o << " attribute labels: " << attribute_labels.size() << std::endl;
			o << " object labels: " << object_labels.size() << std::endl;
			o << " edges: " << edges.size() << std::endl;
		}
	};

	/// output statistics
	/** \relates diagram
	    \param o output stream
	    \param d diagram, whose statistic shall be output
	    \return reference to the output stream
	*/
	inline std::ostream & operator << (std::ostream & o , const diagram & d) {
		return d.statistics(o);
	}
  
	/// list of diagrams
	/* relates diagram */
	typedef std::list<diagram> diagram_s;

	/// appends a diagram to a diagram list
	/** \relates diagram_s
	    \relates diagram
	    \param s list (set) of diagrams
	    \param d diagram to be appended
	    \return reference to the diagram list 
	*/
	inline diagram_s& operator+=(diagram_s& s, diagram d) {
		s.push_back(d);
		return s;
	}

	inline diagram::diagram(const xmlpp::Node * Node,
				context_s & c):node(Node),
					       contexts(c),
					       min_node_radius(2)
	{
		if (!Node) throw xmlpp::validity_error("Empty parent Node for diagrams.");

#ifdef DEBUG
		std::cerr << "Diagram: " << Node -> get_path() << std::endl;
#endif
    
		load_settings(Node);
    
		// loading concepts
		xmlpp::NodeSet nodes = Node->find(CONCEPT_FIGURES_ELEMENT + "/" 
						  + FIGURE);
		for (xmlpp::NodeSet::iterator i = nodes.begin();
		     i != nodes.end(); i++) {
			if (parse_attribute(*i,FIGURE_TYPE_ATTRIBUTE,false) 
			    == CONCEPT_FIGURE_TYPE) 
				concepts += concept(*i);
		}
    
		// loading down labels
		nodes = Node->find(CONCEPT_LABELS_ELEMENT + "/" 
				   + FIGURE);
		for (xmlpp::NodeSet::iterator i = nodes.begin();
		     i != nodes.end(); i++) {
			if (parse_attribute(*i,FIGURE_TYPE_ATTRIBUTE,false) 
			    == CONCEPT_LABEL_FIGURE_TYPE) 
				downlabels += label(*i);
		}
    
		// loading up labels
		nodes = Node->find(UP_CONCEPT_LABELS_ELEMENT + "/" 
				   + FIGURE);
		for (xmlpp::NodeSet::iterator i = nodes.begin();
		     i != nodes.end(); i++) {
			if (parse_attribute(*i,FIGURE_TYPE_ATTRIBUTE,false) 
			    == CONCEPT_LABEL_FIGURE_TYPE) 
				uplabels += label(*i);
		}

		// loading attribute labels
		nodes = Node->find(ATTRIBUTE_LABELS_ELEMENT + "/" 
				   + FIGURE);
		for (xmlpp::NodeSet::iterator i = nodes.begin();
		     i != nodes.end(); i++) {
			if (parse_attribute(*i,FIGURE_TYPE_ATTRIBUTE,false) 
			    == ATTRIBUTE_LABEL_FIGURE_TYPE) 
				attribute_labels += attribute_label(*i);
		}
    
		// loading object labels
		nodes = Node->find(OBJECT_LABELS_ELEMENT + "/" 
				   + FIGURE);
		for (xmlpp::NodeSet::iterator i = nodes.begin();
		     i != nodes.end(); i++) {
			if (parse_attribute(*i,FIGURE_TYPE_ATTRIBUTE,false) 
			    == OBJECT_LABEL_FIGURE_TYPE) 
				object_labels += object_label(*i);
		}
   
		set_concept_objects();
		order.set(concepts);
		find_edges();
		set_concept_extents();
		set_concept_attr();
		node_radius_max_value = get_object_count();
		update_uplabels();
		update_downlabels();
	}

	inline void diagram::load_settings(const xmlpp::Node * Node) {
		if (!Node) throw xmlpp::validity_error("Empty parent Node for diagrams.");

#ifdef DEBUG
		std::cerr << "Diagram: " << Node -> get_path() << std::endl;
#endif

		const Glib::ustring base = "LineDiagramSettings/";
		Glib::ustring tmp ; 

		attribute_label_display_mode = 
			parse_single_node_attribute (Node, 
						     base + "AttributeLabelsDisplayMode");

		object_label_display_mode = 
			parse_single_node_attribute (Node, 
						     base + "ObjectLabelsDisplayMode");
    
		label_font_size = 
			atoi(parse_single_node_attribute (Node,
							  base + "LabelFontSize").c_str());
    
		tmp = parse_single_node_attribute (Node, base + SHOW_COLLISIONS);
		show_collisions = stob(tmp);

		node_radius = 
			Glib::Ascii::strtod(parse_single_node_attribute (Node,
									 base + MAX_NODE_RADIUS).c_str());

		node_radius_mode = 
			parse_single_node_attribute (Node, base + NODE_RADIUS_MODE);
	   
		edge_display_mode = 
			parse_single_node_attribute (Node, base + EDGE_DISPLAY_MODE);

		highlight_mode = 
			parse_single_node_attribute(Node, base + HIGHLIGHT_MODE);

		lattice_layout = 
			parse_single_node_attribute(Node, base + LATTICE_LAYOUT);

		grid_size_x = 
			atoi(parse_single_node_attribute(Node, base + GRID_SIZE_X).c_str());
		grid_size_y = 
			atoi(parse_single_node_attribute(Node, base + GRID_SIZE_Y).c_str());
	}

	inline concept_rs diagram::get_irreducibles() const 
	{
		concept_rs set;
		for (concept_order::const_iterator i=order.begin();i!=order.end();i++) {
			const concept_sibling_rl &siblings = i->get_neighbours();
			if (siblings.size() <= 1)
				set.push_back(concepts.find(i->get_ref()));
		}
		return set;
	}

	inline void diagram::find_edges() {
		for (concept_order::iterator i=order.begin();i!=order.end();i++) {
			const concept_sibling_rl &siblings = (*i).get_neighbours();
			for (concept_sibling_rl::const_iterator j = siblings.begin() ; 
			     j!= siblings.end(); j++) {
				edges += edge(*i,*(*j));
			}
		}
	}

	inline void diagram::set_concept_objects() {
		typedef std::vector<object_s::iterator> objlist;
		objlist start,end;

		// glue together cotext object lists
		for (context_s::iterator c=contexts.begin(); c != contexts.end(); c++) {
			object_s::const_iterator o = (c->second.get_objects().begin());
			start.push_back(o);
			o = c->second.get_objects().end();
			end.push_back(o);
		}
    
		// find object concepts
		bool last = false;
		for (concept_s::iterator c=concepts.begin(); 
		     c!= concepts.end() && !last; 
		     c++) {
			last = true;
			for (int b = 0; b < start.size(); b++) {

				if (start[b] != end[b]) {
					last = false;
					assert(!(*start[b] < *c));
					while (!((*c) < *(start[b]))) {
						*(const_cast<concept *>(&(*c))) += (*(start[b]));
#ifdef DEBUG
						object_s ::iterator x = start[b];
						std::cerr << (Glib::ustring) *start[b] << ": "
							  << start[b]->get_intent() << std::endl;
#endif
						start[b]++;
						if (start[b] == end[b]) break;
#ifdef DEBUG
						assert(!(start[b] == x));
#endif
					}
				}
			}
		}
	}

	inline void diagram::set_concept_extents() {
		order.check_objects();
	}  

	inline double diagram::get_node_radius(const concept & c) const {
		if (node_radius_mode == "MaxNodeRadiusCalcStrategy") {
			return node_radius;
		} else 
			std::cerr << "Warning: unsupported node drawing mode \""
				  << node_radius_mode << "\"." << std::endl;  
		return get_node_radius (c.get_object_count() / node_radius_max_value);
	}
  
	inline void diagram::set_concept_attr() {
		order.check_attributes();
	}  

	inline void diagram::update_downlabels() {
		label_s::iterator i;
		for (i=downlabels.begin(); i!=downlabels.end() ; i++) {
			concept_s::iterator c = concepts.find(i->get_intent());
			i->update_concept(*c);
			i->set_object_names();
		}
	}

	inline void diagram::update_uplabels() {
		label_s::iterator i;
		for (i=uplabels.begin(); i!=uplabels.end() ; i++) {
			concept_s::iterator c = concepts.find(i->get_intent());
			i->update_concept(*c);
			i->set_attribute_names(contexts.begin()->second.get_attributes());
		}
	}
}

#endif /* CEXFILE_DIAGRAM_H */
