/*       -*- C++ -*- */
/** \file
 ********************************************************************
 * $URL: svn+ssh://nalw02/home/alg/schlemme/svn/cextools/trunk/include/cexfile/intent.h $
 * $Id: intent.h 13 2008-04-21 08:52:07Z schlemme $
 * $Date: 2008-04-21 10:52:07 +0200 (Mo, 21 Apr 2008) $
 * $Revision: 13 $
 * \author Tobias Schlemmer <keinstein@users.berlios.de>
 * $Author: schlemme $ 
 * \license LGPL
 * 
 * Intent class.
 ********************************************************************
 */	  
#ifndef CEXFILE_INTENT_H
#define CEXFILE_INTENT_H

#include <iostream>
#include <vector>
#include "cexfile/attribute.h"

namespace cexfile {
  /// an extended set of attribute identifiers or intent
  class intent:public attribute_indices {
    /// XML node the intent belongs to 
    const xmlpp::Node * node;
  public:
    /// default constructor
    intent(): attribute_indices(),node(NULL) {}

    /// standard constructor
    /** \param Node XML node to be parsed 
     */
    intent(const xmlpp::Node * Node);

    /// add an attribute identfier
    /** \param k identifier to be added
	\return self reference 
    */
    intent& operator+=(const attribute_s::key_type k) {
      insert(k);
      return *this;
    }

    /// output the intent to a standard stream
    /** \param o reference to the output stream
	\return reference to the output stream \p o
    */
    std::ostream & print (std::ostream & o) const {
      o << "(intent";
      for (const_iterator i= begin(); i!=end() ; i++) {
	o << " " << *i;
      }
      return o << ")";
    }

    // std::set comparison: bool operator > (onnst intent & i) const; 
    /// linear ordering of intents
    /** \param i other intent to be compared with us
	\retval true if we are size-lexicographically smaller
	\retval false otherwise
    */
    bool operator < (const intent & i) const;

    /// set ordering
    /** \param i intent to be compared with
	\retval true if we are a subset or equal to \p i
	\retval false otherwise
    */
    bool operator <= (const intent & i) const;

    /// set subtraction
    /** \param i items to be removed
	\retval self reference after removing the items from \p i
    */
    intent & operator -= (const intent & i);
  };

  /// output operator
  /** calls intent::print() 
      \relates intent
      \param o output stream
      \param i intent to be sent to the stream
      \return referenct to the output stream
  */
  inline std::ostream & operator << (std::ostream & o, const intent & i) {
    return i.print(o);
  }
  
  inline intent::intent(const xmlpp::Node * Node):attribute_indices(),
						  node(Node)
  {
    if (!Node) throw xmlpp::validity_error("Empty parent Node for intents.");

#ifdef DEBUG
    std::cerr << "Intent: " << Node -> get_path() << std::endl;
#endif
    
    xmlpp::NodeSet att = Node->find("./Intent");
    if (att.size() != 1) 
      throw xmlpp::validity_error
	("Currently there is only one intent supported.");

#ifdef DEBUG
    std::cerr << "Intent: " << (*(att.begin())) -> get_path() << std::endl;
#endif
    
    att = (*(att.begin()))->find("HasAttribute");

#ifdef DEBUG
    std::cerr << "     size: " << att.size() << std::endl;
#endif


    for (xmlpp::NodeSet::iterator i = att.begin(); i != att.end(); i++){

      const xmlpp::Element* nodeElement = CexElementCast(*i);
    
      if (nodeElement) {
	xmlpp::Attribute * Attr = 
	  nodeElement->get_attribute("AttributeIdentifier");

	if (Attr) 
	  insert(Attr -> get_value());
	else throw xmlpp::validity_error("Intent attribute has no Identifier");
      } else 
	throw xmlpp::validity_error
	  ("Intent attribute must have an Identifier.");
    }
  }


  inline bool intent::operator< (const intent& other) const 
  {
    if (size() > other.size()) return false;
    if (size() < other.size()) return true;
    return attribute_indices(*this) < attribute_indices(other);
  }

  
  // subset or equal
  inline bool intent::operator<= (const intent& other) const 
  {
    if (size() > other.size()) return false;
    for (const_iterator i1=begin(),i2=other.begin(); 
	 i1 != end(); i1++, i2++) {
      if (i2 == other.end()) return false;
      while((*i1 > *i2)){
	if (++i2 == other.end()) return false;
      }
      if (*i1 < *i2) return false;
    }
    return true;
  }

  inline intent & intent::operator -= (const intent & i) {
    iterator i2 = i.begin();
    iterator last_i1 = begin();
    if (i2==i.end()) return *this;
#ifdef DEBUG
    std::cerr << "Starting..." << std::endl
	      << "i2: " << *i2 << *this << i << std::endl;
#endif
    for(iterator i1 = begin();
	i1 != end();) {
#ifdef DEBUG
      std::cerr << "i1: " << *i1 << *this << i << std::endl;
#endif
      while (*i2 < *i1) {
	i2++; 
	if (i2 == i.end()) return *this;
#ifdef DEBUG
      std::cerr << "i2: " << *i2 << *this << i << std::endl;
#endif
      }
      if (*i1 == *i2) {
#ifdef DEBUG
	std::cerr << "erasing " << *i1 << *this << i << std::endl;
#endif 	
	
	// guard against some unintended effect since we are 
	// biting the hand that feeds us
	if (i1 == begin()) {
	  CEXDEBUG("we are the first");
	  erase(i1);
	  last_i1 = i1 = begin();
	} else {
	  CEXDEBUG("we are stepping back to " << *last_i1);
	  erase(i1);
	  i1 = last_i1;
	  i1++;
	}
      } else {
	last_i1 = i1;
	i1++;
      }
    }
    return *this;
  }

}

#endif /* CEXFILE_INTENT_H */
